#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static FILE* fi = 0, *fo = 0, *fo2 = 0;
static char* ni = 0, *no = 0, *no2 = 0;
static unsigned char* buf = 0;
static int fs, o, t, s, sc;

static const int vol = 254;
static const int max_sectors = 16;
static const int nib_gap_size = 7;
static const int use_interleave = 1;

/* forward reference */
static int nib_track_write(int track);

static void cleanup()
{
    if(fo)
    {
        fclose(fo);
        fo = 0;
    }

    if(fi)
    {
        fclose(fi);
        fi = 0;
    }

    if(no)
    {
        free(no);
        no=0;
    }

    if(no2)
    {
        free(no2);
        no=0;
    }

    if(buf)
    {
        free(buf);
        buf=0;
    }
}

static int error(const char* fmt, const char* str, int ecode)
{
    fprintf(stderr, "err: ");
    fprintf(stderr, fmt, str);
    fprintf(stderr, "\n");

    cleanup();

    exit(ecode);
}

int main(int argc, char* argv[])
{
    if(argc != 2)
        error("usage: %s <oric_dsk_image>", argv[0], -1);

    ni = argv[1];
    fi = fopen(ni, "rb");
    if(!fi)
        error("can`t open input file: %s\n", ni, -2);

    fseek(fi, 0, SEEK_END);
    fs = ftell(fi);

    if(224256 != fs)
        error("input file `%s` has bad size (it should be exactly 224256 bytes)", ni, -3);

    no = (char*)malloc(strlen(ni)+4+1);
    if(!no)
        error("out of memory (%s)", "malloc failed", -4);

    no2 = (char*)malloc(strlen(ni)+4+1);
    if(!no2)
        error("out of memory (%s)", "malloc failed", -4);

    buf = (unsigned char*)malloc(max_sectors*256);
    if(!buf)
        error("out of memory (%s)", "malloc failed", -5);

    memset(buf, 0xaa, max_sectors*256);

    strcpy(no, ni);
    strcpy(no2, ni);

    strcat(no, ".dsk");
    strcat(no2, ".nib");

    fo = fopen(no, "wb");
    if(!fo)
        error("can`t open output file: %s\n", no, -6);

    fo2 = fopen(no2, "wb");
    if(!fo2)
        error("can`t open output file: %s\n", no2, -6);

    o = 0x100;
    fseek(fi, o, SEEK_SET);

    sc = 0;
    while(o<fs)
    {
        unsigned char c;
        c = fgetc(fi);
        o++;
        if(0xfe == c)
        {
            t = fgetc(fi);
            fgetc(fi);
            s = fgetc(fi);
            fgetc(fi);
            fgetc(fi);
            fgetc(fi);
            o += 6;

            // printf("\rt:%d s:%d", t, s);
            while(o<fs)
            {
                c = fgetc(fi);
                o++;
                if(0xfb == c || 0xf8 == c)
                {
                    fread(buf+256*(s-1), 256, 1, fi);
                    fgetc(fi);
                    fgetc(fi);
                    o += 256 + 2;
                    sc++;
                    break;
                }
            }
            if(max_sectors == sc)
            {
                fwrite(buf, 256, max_sectors, fo);
                nib_track_write(t);
                memset(buf, 0xaa, max_sectors*256);
                sc = 0;
            }
        }
    }

    fprintf(stdout,"done: diskii file: %s\n", no);
    cleanup();
    return 0;
}


/* nib converter */

static const unsigned char trans_table[] =
{
    0x96, 0x97, 0x9A, 0x9B, 0x9D, 0x9E, 0x9F, 0xA6,
    0xA7, 0xAB, 0xAC, 0xAD, 0xAE, 0xAF, 0xB2, 0xB3,
    0xB4, 0xB5, 0xB6, 0xB7, 0xB9, 0xBA, 0xBB, 0xBC,
    0xBD, 0xBE, 0xBF, 0xCB, 0xCD, 0xCE, 0xCF, 0xD3,
    0xD6, 0xD7, 0xD9, 0xDA, 0xDB, 0xDC, 0xDD, 0xDE,
    0xDF, 0xE5, 0xE6, 0xE7, 0xE9, 0xEA, 0xEB, 0xEC,
    0xED, 0xEE, 0xEF, 0xF2, 0xF3, 0xF4, 0xF5, 0xF6,
    0xF7, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, 0xFF
};

static const unsigned char interleave[] =
{
    0x00, 0x07, 0x0E, 0x06, 0x0D, 0x05, 0x0C, 0x04,
    0x0B, 0x03, 0x0A, 0x02, 0x09, 0x01, 0x08, 0x0F
};

/*
 * Encode a 256-unsigned char sector as SECTOR_SIZE disk bytes as follows:
 *
 *  40 sync bytes             : FF FF ...
 *   3 address header bytes   : D5 AA 96
 *   8 address block bytes
 *   3 address trailer bytes  : DE AA EB
 *   6 sync bytes             : FF FF FF FF FF FF
 *   3 nib_disk_buf header bytes      : D5 AA AD
 * 343 nib_disk_buf block bytes
 *   3 nib_disk_buf trailer bytes     : DE AA EB
 * ===
 * 409 bytes         (409 * 16 sectors = $1990)
 */
int nib_sector_write(int t, int sector,
                     unsigned char* in, int in_ofs)
{
    int loop, checksum, prev_value, value;
    int sector_buffer[258];
    value = 0;

    /*
     * Step 1: write 40 sync bytes (0xFF's).
     */
    for (loop = 0; loop < 40; loop++)
        fputc((unsigned char) 0xFF,fo2);

    /*
     * Step 2: write the 3-unsigned char address header (0xD5 0xAA 0x96).
     */
    fputc((unsigned char) 0xD5,fo2);
    fputc((unsigned char) 0xAA,fo2);
    fputc((unsigned char) 0x96,fo2);

    /*
     * Step 3: write the address block. Use 4-and-4 encoding to convert the
     * volume, track and sector and checksum into 2 disk bytes each. The
     * checksum is a simple exclusive OR of the first three values.
     */
    fputc((unsigned char) ((vol >> 1) | 0xAA),fo2);
    fputc((unsigned char) (vol | 0xAA),fo2);
    checksum = vol;
    fputc((unsigned char) ((t >> 1) | 0xAA),fo2);
    fputc((unsigned char) (t | 0xAA),fo2);
    checksum ^= t;
    fputc((unsigned char) ((sector >> 1) | 0xAA),fo2);
    fputc((unsigned char) (sector | 0xAA),fo2);
    checksum ^= sector;
    fputc((unsigned char) ((checksum >> 1) | 0xAA),fo2);
    fputc((unsigned char) (checksum | 0xAA),fo2);

    /*
     * Step 4: write the 3-unsigned char address trailer (0xDE 0xAA 0xEB).
     */
    fputc((unsigned char) (0xDE),fo2);
    fputc((unsigned char) (0xAA),fo2);
    fputc((unsigned char) (0xEB),fo2);

    /*
     * Step 5: write another 6 sync bytes.
     */
    for (loop = 0; loop < 6; loop++)
      fputc((unsigned char) (0xFF),fo2);

    /*
     * Step 6: write the 3-unsigned char nib_disk_buf header.
     */
    fputc((unsigned char) (0xD5),fo2);
    fputc((unsigned char) (0xAA),fo2);
    fputc((unsigned char) (0xAD),fo2);

    /*
     * Step 7: read the next 256-unsigned char sector from the old disk image file,
     * and add two zero bytes to bring the number of bytes up to a multiple
     * of 3.
     */
    for (loop = 0; loop < 256; loop++)
        sector_buffer[loop] = in[loop + in_ofs] & 0xFF;
    sector_buffer[256] = 0;
    sector_buffer[257] = 0;

    /*
     * Step 8: write the first 86 disk bytes of the nib_disk_buf block, which
     * encodes the bottom two bits of each sector unsigned char into six-bit values
     * as follows:
     *
     * disk unsigned char n, bit 0 = sector unsigned char n, bit 1 disk unsigned char n, bit 1 = sector
     * unsigned char n, bit 0 disk unsigned char n, bit 2 = sector unsigned char n + 86, bit 1 disk
     * unsigned char n, bit 3 = sector unsigned char n + 86, bit 0 disk unsigned char n, bit 4 = sector
     * unsigned char n + 172, bit 1 disk unsigned char n, bit 5 = sector unsigned char n + 172, bit 0
     *
     * The scheme allows each pair of bits to be shifted to the right out of
     * the disk unsigned char, then shifted to the left into the sector unsigned char.
     *
     * Before the 6-bit value is translated to a disk unsigned char, it is exclusive
     * ORed with the previous 6-bit value, hence the values written are
     * really a running checksum.
     */
    prev_value = 0;
    for (loop = 0; loop < 86; loop++)
    {
        value = (sector_buffer[loop] & 0x01) << 1;
        value |= (sector_buffer[loop] & 0x02) >> 1;
        value |= (sector_buffer[loop + 86] & 0x01) << 3;
        value |= (sector_buffer[loop + 86] & 0x02) << 1;
        value |= (sector_buffer[loop + 172] & 0x01) << 5;
        value |= (sector_buffer[loop + 172] & 0x02) << 3;
        fputc((unsigned char) (trans_table[value ^ prev_value]),fo2);
        prev_value = value;
    }

    /*
     * Step 9: write the last 256 disk bytes of the nib_disk_buf block, which
     * encodes the top six bits of each sector unsigned char. Again, each value is
     * exclusive ORed with the previous value to create a running checksum
     * (the first value is exclusive ORed with the last value of the
     * previous step).
     */

    for (loop = 0; loop < 256; loop++)
    {
        value = (sector_buffer[loop] >> 2);
        fputc((unsigned char) (trans_table[value ^ prev_value]),fo2);
        prev_value = value;
    }

    /*
     * Step 10: write the last value as the checksum.
     */
    fputc((unsigned char) (trans_table[value]),fo2);

    /*
     * Step 11: write the 3-unsigned char nib_disk_buf trailer.
     */
    fputc((unsigned char) (0xDE),fo2);
    fputc((unsigned char) (0xAA),fo2);
    fputc((unsigned char) (0xEB),fo2);

    return 0;
}

static int nib_track_write(int track)
{
    int in_pos = 0;
    int sector;
    int gap;

    for (sector = 0; sector < max_sectors; sector++)
    {
        in_pos = (use_interleave?interleave[sector]:sector)*256;
        nib_sector_write(track, sector, buf, in_pos);

        gap = nib_gap_size;
        while(gap-->0)
          fputc((unsigned char)(0xFF),fo2);
    }

    return 0;
}

